import {
	beforeAll,
	afterAll,
	afterEach,
	beforeEach,
	describe,
	expect,
	it,
	vi,
} from 'vitest';
import {constants} from 'http2';
import {Backend} from '../../../src/backend';
import {
	application,
	resource,
	validation as v,
	Resource,
	Application,
} from '../../../src/common';
import { autoIncrement } from '../../fixtures';
import {createTestClient, DummyDataSource, TEST_LANGUAGE, TestClient} from '../../utils';
import {DataSource} from '../../../src/backend/data-source';

const PORT = 3000;
const HOST = '127.0.0.1';
const BASE_PATH = '/api';
const ACCEPT = 'application/json';
const ACCEPT_LANGUAGE = 'en';
const ACCEPT_CHARSET = 'utf-8';
const CONTENT_TYPE_CHARSET = 'utf-8';
const CONTENT_TYPE = ACCEPT;

describe('happy path', () => {
	let Piano: Resource;
	let app: Application;
	let dataSource: DataSource;
	let backend: Backend;
	let server: ReturnType<Backend['createHttpServer']>;
	let client: TestClient;

	beforeAll(() => {
		Piano = resource(v.object(
			{
				brand: v.string()
			},
			v.never()
		))
			.name('Piano' as const)
			.route('pianos' as const)
			.id('id' as const, {
				generationStrategy: autoIncrement,
				serialize: (id) => id?.toString() ?? '0',
				deserialize: (id) => Number.isFinite(Number(id)) ? Number(id) : 0,
				schema: v.number(),
			});

		app = application({
			name: 'piano-service',
		})
			.language(TEST_LANGUAGE)
			.resource(Piano);

		dataSource = new DummyDataSource();

		backend = app.createBackend({
			dataSource,
		});

		server = backend.createHttpServer({
			basePath: BASE_PATH
		});

		client = createTestClient({
			host: HOST,
			port: PORT,
		})
			.acceptMediaType(ACCEPT)
			.acceptLanguage(ACCEPT_LANGUAGE)
			.acceptCharset(ACCEPT_CHARSET)
			.contentType(CONTENT_TYPE)
			.contentCharset(CONTENT_TYPE_CHARSET);

		return new Promise((resolve, reject) => {
			server.on('error', (err) => {
				reject(err);
			});

			server.on('listening', () => {
				resolve();
			});

			server.listen({
				port: PORT
			});
		});
	});

	afterAll(() => new Promise<void>((resolve, reject) => {
		server.close((err) => {
			if (err) {
				reject(err);
			}

			resolve();
		});
	}));

	describe('serving collections', () => {
		beforeEach(() => {
			vi
				.spyOn(DummyDataSource.prototype, 'getMultiple')
				.mockResolvedValueOnce([] as never);
		});

		beforeEach(() => {
			Piano.canFetchCollection();
		});

		afterEach(() => {
			Piano.canFetchCollection(false);
		});

		it('returns data', async () => {
			const [res, resData] = await client({
				method: 'GET',
				path: `${BASE_PATH}/pianos`,
			});

			expect(res).toHaveProperty('statusCode', constants.HTTP_STATUS_OK);
			expect(res).toHaveProperty('statusMessage', TEST_LANGUAGE.statusMessages.resourceCollectionFetched);
			expect(res.headers).toHaveProperty('content-type', expect.stringContaining(ACCEPT));

			if (typeof resData === 'undefined') {
				expect.fail('Response body must be defined.');
				return;
			}

			expect(resData).toEqual([]);
		});

		it('returns data on HEAD method', async () => {
			const [res] = await client({
				method: 'HEAD',
				path: `${BASE_PATH}/pianos`,
			});

			expect(res).toHaveProperty('statusCode', constants.HTTP_STATUS_OK);
			expect(res).toHaveProperty('statusMessage', TEST_LANGUAGE.statusMessages.resourceCollectionFetched);
		});

		it('returns options', async () => {
			const [res] = await client({
				method: 'OPTIONS',
				path: `${BASE_PATH}/pianos`,
			});

			expect(res).toHaveProperty('statusCode', constants.HTTP_STATUS_NO_CONTENT);
			expect(res).toHaveProperty('statusMessage', TEST_LANGUAGE.statusMessages.provideOptions);
			const allowedMethods = res.headers.allow?.split(',').map((s) => s.trim()) ?? [];
			expect(allowedMethods).toContain('GET');
			expect(allowedMethods).toContain('HEAD');
		});
	});

	describe('serving items', () => {
		const existingResource = {
			id: 1,
			brand: 'Yamaha'
		};

		beforeEach(() => {
			vi
				.spyOn(DummyDataSource.prototype, 'getById')
				.mockResolvedValueOnce(existingResource as never);
		});

		beforeEach(() => {
			Piano.canFetchItem();
		});

		afterEach(() => {
			Piano.canFetchItem(false);
		});

		it('returns data', async () => {
			const [res, resData] = await client({
				method: 'GET',
				path: `${BASE_PATH}/pianos/${existingResource.id}`,
			});

			expect(res).toHaveProperty('statusCode', constants.HTTP_STATUS_OK);
			expect(res).toHaveProperty('statusMessage', TEST_LANGUAGE.statusMessages.resourceFetched);
			expect(res.headers).toHaveProperty('content-type', expect.stringContaining(ACCEPT));
			if (typeof resData === 'undefined') {
				expect.fail('Response body must be defined.');
				return;
			}

			expect(resData).toEqual(existingResource);
		});

		it('returns data on HEAD method', async () => {
			const [res] = await client({
				method: 'HEAD',
				path: `${BASE_PATH}/pianos/${existingResource.id}`,
			});

			expect(res).toHaveProperty('statusCode', constants.HTTP_STATUS_OK);
			expect(res).toHaveProperty('statusMessage', TEST_LANGUAGE.statusMessages.resourceFetched);
		});

		it('returns options', async () => {
			const [res] = await client({
				method: 'OPTIONS',
				path: `${BASE_PATH}/pianos/${existingResource.id}`,
			});

			expect(res).toHaveProperty('statusCode', constants.HTTP_STATUS_NO_CONTENT);
			expect(res).toHaveProperty('statusMessage', TEST_LANGUAGE.statusMessages.provideOptions);
			const allowedMethods = res.headers.allow?.split(',').map((s) => s.trim()) ?? [];
			expect(allowedMethods).toContain('GET');
			expect(allowedMethods).toContain('HEAD');
		});
	});

	describe('creating items', () => {
		const newResourceData = {
			brand: 'K. Kawai'
		};

		const responseData = {
			id: 2,
			...newResourceData,
		};

		beforeEach(() => {
			vi
				.spyOn(DummyDataSource.prototype, 'newId')
				.mockResolvedValueOnce(responseData.id as never);
		});

		beforeEach(() => {
			vi
				.spyOn(DummyDataSource.prototype, 'create')
				.mockResolvedValueOnce(responseData as never);
		});

		beforeEach(() => {
			Piano.canCreate();
		});

		afterEach(() => {
			Piano.canCreate(false);
		});

		it('returns data', async () => {
			const [res, resData] = await client({
				path: `${BASE_PATH}/pianos`,
				method: 'POST',
				body: newResourceData,
			});

			expect(res).toHaveProperty('statusCode', constants.HTTP_STATUS_CREATED);
			expect(res).toHaveProperty('statusMessage', TEST_LANGUAGE.statusMessages.resourceCreated);
			expect(res.headers).toHaveProperty('content-type', expect.stringContaining(ACCEPT));
			expect(res.headers).toHaveProperty('location', `${BASE_PATH}/pianos/2`);

			if (typeof resData === 'undefined') {
				expect.fail('Response body must be defined.');
				return;
			}

			expect(resData).toEqual({
				...newResourceData,
				id: 2
			});
		});

		it('returns options', async () => {
			const [res] = await client({
				method: 'OPTIONS',
				path: `${BASE_PATH}/pianos`,
			});

			expect(res).toHaveProperty('statusCode', constants.HTTP_STATUS_NO_CONTENT);
			expect(res).toHaveProperty('statusMessage', TEST_LANGUAGE.statusMessages.provideOptions);
			const allowedMethods = res.headers.allow?.split(',').map((s) => s.trim()) ?? [];
			expect(allowedMethods).toContain('POST');
		});
	});

	describe('patching items', () => {
		const existingResource = {
			id: 1,
			brand: 'Yamaha'
		};

		const patchData = {
			brand: 'K. Kawai'
		};

		beforeEach(() => {
			vi
				.spyOn(DummyDataSource.prototype, 'getById')
				.mockResolvedValueOnce(existingResource as never);
		});

		beforeEach(() => {
			vi
				.spyOn(DummyDataSource.prototype, 'patch')
				.mockResolvedValueOnce({
					...existingResource,
					...patchData,
				} as never);
		});

		beforeEach(() => {
			Piano.canPatch();
		});

		afterEach(() => {
			Piano.canPatch(false);
		});

		it('returns options', async () => {
			const [res] = await client({
				method: 'OPTIONS',
				path: `${BASE_PATH}/pianos/${existingResource.id}`,
			});

			expect(res).toHaveProperty('statusCode', constants.HTTP_STATUS_NO_CONTENT);
			expect(res).toHaveProperty('statusMessage', TEST_LANGUAGE.statusMessages.provideOptions);
			const allowedMethods = res.headers.allow?.split(',').map((s) => s.trim()) ?? [];
			expect(allowedMethods).toContain('PATCH');
			const acceptPatch = res.headers['accept-patch']?.split(',').map((s) => s.trim()) ?? [];
			expect(acceptPatch).toContain('application/json-patch+json');
			expect(acceptPatch).toContain('application/merge-patch+json');
		});

		describe('on merge', () => {
			beforeEach(() => {
				Piano.canPatch(false).canPatch(['merge']);
			});

			it('returns data', async () => {
				const [res, resData] = await client({
					method: 'PATCH',
					path: `${BASE_PATH}/pianos/${existingResource.id}`,
					body: patchData,
					headers: {
						'content-type': 'application/merge-patch+json',
					},
				});

				expect(res).toHaveProperty('statusCode', constants.HTTP_STATUS_OK);
				expect(res).toHaveProperty('statusMessage', TEST_LANGUAGE.statusMessages.resourcePatched);
				expect(res.headers).toHaveProperty('content-type', expect.stringContaining(ACCEPT));

				if (typeof resData === 'undefined') {
					expect.fail('Response body must be defined.');
					return;
				}

				expect(resData).toEqual({
					...existingResource,
					...patchData,
				});
			});
		});

		describe('on delta', () => {
			beforeEach(() => {
				Piano.canPatch(false).canPatch(['delta']);
			});

			it('returns data', async () => {
				const [res, resData] = await client({
					method: 'PATCH',
					path: `${BASE_PATH}/pianos/${existingResource.id}`,
					body: [
						{
							op: 'replace',
							path: 'brand',
							value: patchData.brand,
						},
					],
					headers: {
						'content-type': 'application/json-patch+json',
					},
				});

				expect(res).toHaveProperty('statusCode', constants.HTTP_STATUS_OK);
				expect(res).toHaveProperty('statusMessage', TEST_LANGUAGE.statusMessages.resourcePatched);
				expect(res.headers).toHaveProperty('content-type', expect.stringContaining(ACCEPT));

				if (typeof resData === 'undefined') {
					expect.fail('Response body must be defined.');
					return;
				}

				expect(resData).toEqual({
					...existingResource,
					...patchData,
				});
			});
		});
	});

	describe('emplacing items', () => {
		const existingResource = {
			id: 1,
			brand: 'Yamaha'
		};

		const emplaceResourceData = {
			id: 1,
			brand: 'K. Kawai'
		};

		beforeEach(() => {
			Piano.canEmplace();
		});

		afterEach(() => {
			Piano.canEmplace(false);
		});

		it('returns data for replacement', async () => {
			vi
				.spyOn(DummyDataSource.prototype, 'emplace')
				.mockResolvedValueOnce([{
					...existingResource,
					...emplaceResourceData,
				}, false] as never);

			const [res, resData] = await client({
				method: 'PUT',
				path: `${BASE_PATH}/pianos/${emplaceResourceData.id}`,
				body: emplaceResourceData,
			});

			expect(res).toHaveProperty('statusCode', constants.HTTP_STATUS_OK);
			expect(res).toHaveProperty('statusMessage', TEST_LANGUAGE.statusMessages.resourceReplaced);
			expect(res.headers).toHaveProperty('content-type', expect.stringContaining(ACCEPT));

			if (typeof resData === 'undefined') {
				expect.fail('Response body must be defined.');
				return;
			}

			expect(resData).toEqual(emplaceResourceData);
		});

		it('returns data for creation', async () => {
			const newId = 2;

			vi
				.spyOn(DummyDataSource.prototype, 'emplace')
				.mockResolvedValueOnce([{
					...existingResource,
					...emplaceResourceData,
					id: newId
				}, true] as never);

			const [res, resData] = await client({
				method: 'PUT',
				path: `${BASE_PATH}/pianos/${newId}`,
				body: {
					...emplaceResourceData,
					id: newId,
				},
			});

			expect(res).toHaveProperty('statusCode', constants.HTTP_STATUS_CREATED);
			expect(res).toHaveProperty('statusMessage', TEST_LANGUAGE.statusMessages.resourceCreated);
			expect(res.headers).toHaveProperty('content-type', expect.stringContaining(ACCEPT));
			expect(res.headers).toHaveProperty('location', `${BASE_PATH}/pianos/${newId}`);

			if (typeof resData === 'undefined') {
				expect.fail('Response body must be defined.');
				return;
			}

			expect(resData).toEqual({
				...emplaceResourceData,
				id: newId,
			});
		});

		it('returns options', async () => {
			const [res] = await client({
				method: 'OPTIONS',
				path: `${BASE_PATH}/pianos/${existingResource.id}`,
			});

			expect(res).toHaveProperty('statusCode', constants.HTTP_STATUS_NO_CONTENT);
			expect(res).toHaveProperty('statusMessage', TEST_LANGUAGE.statusMessages.provideOptions);
			const allowedMethods = res.headers.allow?.split(',').map((s) => s.trim()) ?? [];
			expect(allowedMethods).toContain('PUT');
		});
	});

	describe('deleting items', () => {
		const existingResource = {
			id: 1,
			brand: 'Yamaha'
		};

		beforeEach(() => {
			vi
				.spyOn(DummyDataSource.prototype, 'getById')
				.mockResolvedValueOnce(existingResource as never);
		});

		beforeEach(() => {
			vi
				.spyOn(DummyDataSource.prototype, 'delete')
				.mockReturnValueOnce(Promise.resolve() as never);
		});

		beforeEach(() => {
			Piano.canDelete();
		});

		afterEach(() => {
			Piano.canDelete(false);
		});

		it('responds', async () => {
			const [res, resData] = await client({
				method: 'DELETE',
				path: `${BASE_PATH}/pianos/${existingResource.id}`,
			});

			expect(res).toHaveProperty('statusCode', constants.HTTP_STATUS_NO_CONTENT);
			expect(res).toHaveProperty('statusMessage', TEST_LANGUAGE.statusMessages.resourceDeleted);
			expect(res.headers).not.toHaveProperty('content-type');
			expect(resData).toBeUndefined();
		});

		it('returns options', async () => {
			const [res] = await client({
				method: 'OPTIONS',
				path: `${BASE_PATH}/pianos/${existingResource.id}`,
			});

			expect(res).toHaveProperty('statusCode', constants.HTTP_STATUS_NO_CONTENT);
			expect(res).toHaveProperty('statusMessage', TEST_LANGUAGE.statusMessages.provideOptions);
			const allowedMethods = res.headers.allow?.split(',').map((s) => s.trim()) ?? [];
			expect(allowedMethods).toContain('DELETE');
		});
	});
});
